(*
** 7/31/96 - Created by calix@together.net
**
** Just do the following :
** 1. add a panel to the form,
** 2. add the TSnapToGrid to the form,
** 3. assign the panel to the SnapControl property,
** 4. drop more panels inside the first panel,
** 5. run the form,
** 6. move/resize the contained panels.

** TSnapToGrid is a control with a SnapControl property.
** When the SnapControl property is set, the user can
** move all *WINDOWED* decendents of SnapControl at runtime.
**
** This control *ONLY* allows the user to move controls
** that decend from TWinControl and that are contained
** by SnapControl.
**
** I obtained all the information required to create this
** component from the Delphi32 source code and online help.
**
** 8/1/96 - Added control resizing
** 8/4/96 - Added DrawFocusRect property
** 8/11/96 - Added OnSelect event
** 8/18/96 - Changed OnSelect to OnMouseDown event
** 8/18/96 - Added OnMouseUp event
** 8/18/96 - Added OnMouseMove event
** 9/3/96 - Added Resizeable and Moveable properties
*)

unit cSnapToGrid;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TMoveEvent = procedure (Sender: TObject; Shift: TShiftState; X, Y: Integer) of object;
  TGridSize = SmallInt;
  TSnapToGrid = class(TComponent)
  private
    FWinControl : TWinControl;
    FGridSize : TGridSize;
    FHitSize : Integer;
    FDrawFocusRect : Boolean;
    FMoveable,
    FResizeable : Boolean;

    FOnMouseDown,
    FOnMouseUp : TMouseEvent;
    FOnMouseMove : TMoveEvent;

    OldAppMessage: TMessageEvent;
    intHitTest : Integer;

    ScreenDC : HDC;
    rctOriginal,
    rctMove : TRect;

    ptCapture,
    ptWorld: TPoint;
    bolCapture: Boolean;
    ctrActive: TControl;

    procedure SetGridSize(grdSize: TGridSize);
    procedure HandleMouseMessage(var Msg: TMsg; var Handled: Boolean);
    procedure HandleKeyboardMessage(var Msg: TMsg; var Handled: Boolean);
    procedure HandleGridMessage(var Msg: TMsg; var Handled: Boolean);

  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;

    procedure ChildMouseMove(Shift: TShiftState; X, Y: Integer); virtual;
    procedure ChildMouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); virtual;
    procedure ChildMouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); virtual;
    procedure CTDrawFocusRect;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AppMessage(var Msg: TMsg; var Handled: Boolean);
    { should be CTSnap }
    function Snap(intValue: Integer): Integer;

  published
    //property SnapControl: TWinControl read FWinControl write FWinControl; //obsolete
    property SnapToGridWindow: TWinControl read FWinControl write FWinControl;
    property GridSize: TGridSize read FGridSize write SetGridSize;
    //property FocusRect: Boolean read FDrawFocusRect write FDrawFocusRect; // obsolete
    property isFocusRect: Boolean read FDrawFocusRect write FDrawFocusRect;
    property isMoveable: Boolean read FMoveable write FMoveable;
    property isResizeable: Boolean read FResizeable write FResizeable;

    property OnMouseDown: TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseUp: TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnMouseMove: TMoveEvent read FOnMouseMove write FOnMouseMove;
  end;

var
  msgLast : TMessage;

procedure Register;
function GetControl(Parent: TWinControl; Hwnd: Hwnd): TControl;

implementation

{-------------------------------------------------------}
function TSnapToGrid.Snap(intValue: Integer): Integer;
begin
  // the -FGridSize works because of two's compliment encoding
  // and the fact that FGridSize is always a power of two;
  // otherwise, you'd need something like this :
  // Result := FGridSize *((intValue + (FGridSize div 2)) div FGridSize);
  { // this is a faster way
  if (intValue < 0)
    then Result := -((-intValue + (FGridSize shr 1)) AND -FGridSize)
    else Result := (intValue + (FGridSize shr 1)) AND -FGridSize;
  }
  // using this for simplicity
  Result := FGridSize * Round(intValue / FGridSize);
end;

{-------------------------------------------------------}
procedure TSnapToGrid.CTDrawFocusRect;
var
  rctFocus: TRect;
begin
  rctFocus.TopLeft := SnapToGridWindow.ClientToScreen(Point(0, 0));
  rctFocus.BottomRight := rctFocus.TopLeft;
  with rctMove do begin
    inc(rctFocus.Top, Top);
    inc(rctFocus.Left, Left);
    inc(rctFocus.Bottom, Bottom);
    inc(rctFocus.Right, Right);
  end;
  DrawFocusRect(ScreenDC, rctFocus);
end;

{-------------------------------------------------------}
constructor TSnapToGrid.Create(AOwner: TComponent);
begin
  inherited;
  { initialize }
  FGridSize := 8;
  FHitSize := 8;
  FDrawFocusRect := FALSE;
  FMoveable := TRUE;
  FResizeable := TRUE;

  { get old OnMessage and process after were done with it }
  OldAppMessage := Application.OnMessage;
  Application.OnMessage := AppMessage;
end;

{-------------------------------------------------------}
destructor TSnapToGrid.Destroy;
begin
  Application.OnMessage := OldAppMessage;
  inherited;
end;

{-------------------------------------------------------}
(*
** make sure grdSize is 0 or a power of 2
** grid size of 0 -> disable moving
*)
procedure TSnapToGrid.SetGridSize(grdSize: TGridSize);
var
  i: Integer;
begin
  if grdSize <= 0 then FGridSize := 0
  else begin
    i := 0;
    while ((grdSize shr i) > 1) do inc(i);
    FGridSize := (1 shl i);
  end;
end;

{-------------------------------------------------------}
(*
** Application.OnMessage calls this method.
** This method determines if the message is relevent.
** If it is, it passes it to HandleGridMessage.
*)
procedure TSnapToGrid.AppMessage(var Msg: TMsg; var Handled: Boolean);
begin
  if Assigned(FWinControl) then begin
    ctrActive := GetControl(FWinControl, Msg.Hwnd);
    if Assigned(ctrActive) then begin
      HandleGridMessage(Msg, Handled);
    end;
  end;
  if not(Handled) AND (Assigned(OldAppMessage)) then OldAppMessage(Msg, Handled);
end;

{-------------------------------------------------------}
(*
** Determines the type of message that is being handled and
** takes the necessary steps.
*)
procedure TSnapToGrid.HandleGridMessage(var Msg: TMsg; var Handled: Boolean);
begin
  if (FWinControl <> ctrActive) then begin
    case Msg.message of
      WM_MOUSEMOVE..WM_LBUTTONUP : HandleMouseMessage(Msg, Handled);
      //WM_SYSKEYDOWN : HandleKeyboardMessage(Msg, Handled);
    end; {case }
  end;
end;

{-------------------------------------------------------}
(*
** A mouse event occured.
** Determine what kind of event it was and pass it along to
** the appropriate message handlers.
*)
procedure TSnapToGrid.HandleMouseMessage(var Msg: TMsg; var Handled: Boolean);
var
  Shift : TShiftState;
  X, Y : Integer;
  Button : TMouseButton;
begin
  //Handled := true;

  case (Msg.message) of
    //WM_MOUSEMOVE
    WM_LBUTTONDOWN..WM_LBUTTONDBLCLK: Button := mbLeft;
    WM_RBUTTONDOWN..WM_RBUTTONDBLCLK: Button := mbRight;
    WM_MBUTTONDOWN..WM_MBUTTONDBLCLK: Button := mbMiddle;
    else Button := mbLeft;
  end;

  Shift := [];
  if (MK_CONTROL = (MK_CONTROL and Msg.wparam)) then Shift := Shift + [ssCtrl];
  if (MK_LBUTTON = (MK_LBUTTON and Msg.wparam)) then Shift := Shift + [ssLeft];
  if (MK_MBUTTON = (MK_MBUTTON and Msg.wparam)) then Shift := Shift + [ssMiddle];
  if (MK_RBUTTON = (MK_RBUTTON and Msg.wparam)) then Shift := Shift + [ssRight];
  if (MK_SHIFT = (MK_SHIFT and Msg.wparam)) then Shift := Shift + [ssShift];

  X := LOWORD(Msg.lparam);
  Y := HIWORD(Msg.lparam);

  case Msg.message of
    WM_MOUSEMOVE :
    begin
      ChildMouseMove(Shift, X, Y);
    end;
    WM_LBUTTONDOWN :
    begin
      SetCapture(Msg.Hwnd);
      //ScreenDC := GetDC(FWinControl.Handle);
      ScreenDC := GetDC(0);
      ChildMouseDown(Button, Shift, X, Y);
    end;
    WM_LBUTTONUP :
    begin
      ChildMouseUp(Button, Shift, X, Y);
      ReleaseCapture;
      //ReleaseDC(FWinControl.Handle, ScreenDC);
      ReleaseDC(0, ScreenDC);
    end;
    {case} else Handled := false;
  end;
end;

{-------------------------------------------------------}
(*
** A keyboard event occured.
** Determine what kind of event it was and pass it along to
** the appropriate message handlers.
*)
procedure TSnapToGrid.HandleKeyboardMessage(var Msg: TMsg; var Handled: Boolean);
begin
{}
end;

{------------------------------------------------}
{ excuse me for not declaring constants :
        1 = Left
        2 = Right
        4 = Top
        8 = Bottom
        1+4 = 5 = TopLeft
        etc..
}
procedure TSnapToGrid.ChildMouseMove(Shift: TShiftState; X, Y: Integer);

  procedure CaptureOn;
  var
    dLeft, dWidth, dTop, dHeight : Integer;
  begin
      ptWorld := ctrActive.ClientToScreen(Point(X, Y));

      // reset offsets
      dLeft := 0; dWidth := 0; dTop := 0; dHeight := 0;

      // if we're inside the control
      // then we're dragging
      if (intHitTest = 0) then begin
        if not (FMoveable) then exit;
        dLeft := (ptWorld.X-ptCapture.X);
        dTop := (ptWorld.Y-ptCapture.Y);

      // if we're on the border of the control
      // then we're resizing
      end else begin
        // calculate horizontal offsets
        if 0 <> (intHitTest AND 1) then begin
          dLeft := (ptWorld.X-ptCapture.X);
          dWidth := -dLeft;
        end else if 0 <> (intHitTest AND 2) then begin
          dWidth := (ptWorld.X-ptCapture.X);
        end;

        // calculate vertical offsets
        if 0 <> (intHitTest AND 4) then begin
          dTop := (ptWorld.Y-ptCapture.Y);
          dHeight := -dTop;
        end else if 0 <> (intHitTest AND 8) then begin
          dHeight := (ptWorld.Y-ptCapture.Y);
        end;

      end; { hittest <> 0 }

      // move or resize object
      begin
        if not(ssCtrl in Shift) then begin
          dLeft := Snap(dLeft);
          dWidth := Snap(dWidth);
          dTop := Snap(dTop);
          dHeight := Snap(dHeight);
        end;

        if (FDrawFocusRect) then CTDrawFocusRect;
        with rctOriginal do begin
          rctMove := Bounds(Left+dLeft,Top+dTop,Right+dWidth,Bottom+dHeight);
        end;
        if (FDrawFocusRect) then CTDrawFocusRect
        else ctrActive.BoundsRect := rctMove;
      end;
  end;

  procedure CaptureOff;
  begin
    intHitTest := 0;
    with ctrActive do begin
      if (ssShift in Shift) OR NOT(FResizeable) then begin
        Cursor := crDefault;
        exit;
      end;
      // are we on the border
      if (Width-X >  X)
        then if (X < FHitSize) then intHitTest := intHitTest OR 1 else //LEFT
        else if (X+FHitSize > Width) then intHitTest := intHitTest OR 2; // RIGHT
      if (Height-Y > Y)
        then if (Y < FHitSize) then intHitTest := intHitTest OR 4 else // TOP
        else if (Y+FHitSize > Height) then intHitTest := intHitTest OR 8; // BOT

      case intHitTest of
        0 : Cursor := crDefault;
        1 : Cursor := crSizeWE;
        2 : Cursor := crSizeWE;
        4 : Cursor := crSizeNS;
        5 : Cursor := crSizeNWSE;
        6 : Cursor := crSizeNESW;
        8 : Cursor := crSizeNS;
        9 : Cursor := crSizeNESW;
        10: Cursor := crSizeNWSE;
      end;
    end;
  end;

// beginning of ChildMouseMove
begin
  if Assigned(FOnMouseMove) then FOnMouseMove(ctrActive, Shift, X, Y);
  if (bolCapture) then CaptureOn else CaptureOff;
end;

{------------------------------------------------}
procedure TSnapToGrid.ChildMouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnMouseDown) then FOnMouseDown(ctrActive, Button, Shift, X, Y);
  bolCapture := true;
  with ctrActive do begin
    ptCapture := ClientToScreen(Point(X, Y));
    rctOriginal := Rect(Left, Top, Width, Height);
    rctMove := BoundsRect;
  end;
  if (FDrawFocusRect) then CTDrawFocusRect;
end;

{------------------------------------------------}
procedure TSnapToGrid.ChildMouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Assigned(FOnMouseUp) then FOnMouseUp(ctrActive, Button, Shift, X, Y);
  bolCapture := false;
  if (FDrawFocusRect) then begin
    CTDrawFocusRect;
  end;
  ctrActive.BoundsRect := rctMove;
end;

{-------------------------------------------------------}
procedure TSnapToGrid.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited;
  if (AComponent = FWinControl) AND (Operation = opRemove) then FWinControl := nil;
end;

{-------------------------------------------------------}
(*
** Given a Parent and a handle to a window, determine
** if the window is a descendent of the Parent.
** If it is, return the child control.
*)
function GetControl(Parent: TWinControl; Hwnd: Hwnd): TControl;
var
  i: Integer;
begin
  result := nil;
  if (Parent.handle = Hwnd) then Result := Parent
  else for i := 0 to Parent.ControlCount-1 do begin
    // check all window control descendents
    if (Parent.Controls[i] is TWinControl) then begin
      if (TWinControl(Parent.Controls[i]).Handle = Hwnd)
        then Result := Parent.Controls[i]
        // comment the next line to disable searching indirect descendants
        ;//else Result := GetControl(TWinControl(Parent.Controls[i]), Hwnd);
      if (Result <> nil) then exit;
    end;
  end;
end;


{-------------------------------------------------------}
procedure Register;
begin
  RegisterComponents('Coretech', [TSnapToGrid]);
end;

end.
